/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "PaintHTMLText.h"

#include "HTMLDrawer.h"

#include "SettingsHelpers.h"
#include "PaintHelpers.h"

//Text Alignment
//-----------------------------------------------------
#define ALIGN_LEFT 0
#define ALIGN_TOP 0
#define ALIGN_CENTER 1
#define ALIGN_RIGHT 2
#define ALIGN_BOTTOM 2
//-----------------------------------------------------

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

PaintHTMLText::PaintHTMLText() :
	m_bUseAlphaMap(false),

	m_bFontFade(false),


	m_bFontVisible(true),

	m_iFontAlphaTransparency(255),	

	m_strBorders(""),
	m_iLeftBorder(0),
	m_iTopBorder(0),
	m_iRightBorder(0),
	m_iBottomBorder(0),

	m_strFontName("Arial"),
	m_iFontHeight(15),
	m_clrFontColor(RGB(0,0,0)),

	m_shFontCharSet(DEFAULT_CHARSET),

	m_bFontSmoothing(true),
	m_bFontClearType(true),

	m_bFontBold(false),
	m_bFontItalic(false),
	m_bFontUnderline(false),

	m_bFontShadow(false),
	m_iFontShadowX(1),
	m_iFontShadowY(1),
	m_iFontBlockedShadow(0),
	m_clrFontShadowColor(RGB(128,128,128)),

	m_bFontOutline(false),
	m_clrFontOutlineColor(RGB(128,128,128)),
	m_bFontOutlineExtra(false),

	m_clrFontEmbossColor1(RGB(255,0,255)),
	m_clrFontEmbossColor2(RGB(255,0,255)),

	m_iFontAlign(DT_CENTER),
	m_iFontVertAlign(ALIGN_CENTER),

	m_bFontNoEllipsis(false),

	m_strFontFadeCFG("0"),
	m_iLeftFade(0),
	m_iRightFade(0),
	m_iTopFade(0),
	m_iBottomFade(0),

	m_bLineBreak(false),
	m_iNewlineSpace(0)
{
}

PaintHTMLText::PaintHTMLText(const PaintTextDefaults *pPaintTextDefaults) :
	m_bUseAlphaMap(false),

	m_bFontFade(false)
{
	m_bFontVisible = (pPaintTextDefaults->m_bFontVisible ? true : false);

	m_iFontAlphaTransparency = pPaintTextDefaults->m_iFontAlphaTransparency;

	m_strBorders = pPaintTextDefaults->m_strBorders;
	m_iLeftBorder = pPaintTextDefaults->m_iLeftBorder;
	m_iTopBorder = pPaintTextDefaults->m_iTopBorder;
	m_iRightBorder = pPaintTextDefaults->m_iRightBorder;
	m_iBottomBorder = pPaintTextDefaults->m_iBottomBorder;

	m_strFontName = pPaintTextDefaults->m_strFontName;
	m_iFontHeight = pPaintTextDefaults->m_iFontHeight;
	m_clrFontColor = pPaintTextDefaults->m_clrFontColor;

	m_shFontCharSet = pPaintTextDefaults->m_shFontCharSet;

	m_bFontSmoothing = (pPaintTextDefaults->m_bFontSmoothing ? true : false);
	m_bFontClearType = (pPaintTextDefaults->m_bFontClearType ? true : false);

	m_bFontBold = (pPaintTextDefaults->m_bFontBold ? true : false);
	m_bFontItalic = (pPaintTextDefaults->m_bFontItalic ? true : false);
	m_bFontUnderline = (pPaintTextDefaults->m_bFontUnderline ? true : false);

	m_bFontShadow = (pPaintTextDefaults->m_bFontShadow ? true : false);
	m_iFontShadowX = pPaintTextDefaults->m_iFontShadowX;
	m_iFontShadowY = pPaintTextDefaults->m_iFontShadowY;
	m_iFontBlockedShadow = pPaintTextDefaults->m_iFontBlockedShadow;
	m_clrFontShadowColor = pPaintTextDefaults->m_clrFontShadowColor;

	m_bFontOutline = (pPaintTextDefaults->m_bFontOutline ? true : false);
	m_clrFontOutlineColor = pPaintTextDefaults->m_clrFontOutlineColor;
	m_bFontOutlineExtra = (pPaintTextDefaults->m_bFontOutlineExtra ? true : false);

	m_clrFontEmbossColor1 = pPaintTextDefaults->m_clrFontEmbossColor1;
	m_clrFontEmbossColor2 = pPaintTextDefaults->m_clrFontEmbossColor2;

	m_iFontAlign = pPaintTextDefaults->m_iFontAlign;
	m_iFontVertAlign = pPaintTextDefaults->m_iFontVertAlign;

	m_bFontNoEllipsis = (pPaintTextDefaults->m_bFontNoEllipsis ? true : false);

	m_strFontFadeCFG = pPaintTextDefaults->m_strFontFadeCFG;
	m_iLeftFade = pPaintTextDefaults->m_iLeftFade;
	m_iRightFade = pPaintTextDefaults->m_iRightFade;
	m_iTopFade = pPaintTextDefaults->m_iTopFade;
	m_iBottomFade = pPaintTextDefaults->m_iBottomFade;

	m_bLineBreak = (pPaintTextDefaults->m_bLineBreak ? true : false);
	m_iNewlineSpace = pPaintTextDefaults->m_iNewlineSpace;
}

PaintHTMLText::~PaintHTMLText()
{
}

BOOL PaintHTMLText::configure(LPCSTR pszName, LPCSTR pszPrefix, BOOL bUseAlphaMap, const PaintHTMLText *pDefaultPaintHTMLText, const PaintTextDefaults *pPaintTextDefaults)
{
	string name(pszName);
	string prefix(pszPrefix);

	string strAddToGroup = iGetxRCString(name, (prefix + "FontAddToGroup"), "");

	if (!strAddToGroup.empty() && _stricmp(name.c_str(), strAddToGroup.c_str()) != NULL)
	{	
		return configure(strAddToGroup.c_str(), "", bUseAlphaMap, pDefaultPaintHTMLText, pPaintTextDefaults);
	}

	bool bDifferenceToDefault = false;

	const NameValuePair alignValues[] =
	{
		{ ".left", DT_LEFT },
		{ ".center", DT_CENTER },
		{ ".right", DT_RIGHT },
		{ "left", DT_LEFT },
		{ "center", DT_CENTER },
		{ "right", DT_RIGHT },
		{ 0, 0 }
	};

	const NameValuePair vertAlignValues[] =
	{
		{ ".top", ALIGN_TOP },
		{ ".center", ALIGN_CENTER },
		{ ".bottom", ALIGN_BOTTOM },
		{ "top", ALIGN_TOP },
		{ "center", ALIGN_CENTER },
		{ "bottom", ALIGN_BOTTOM },
		{ 0, 0 }
	};

	const NameValuePair CharSetConstants[] =
	{
		{ "ANSI_CHARSET", ANSI_CHARSET },
		{ "BALTIC_CHARSET", BALTIC_CHARSET },
		{ "CHINESEBIG5_CHARSET", CHINESEBIG5_CHARSET },
		{ "DEFAULT_CHARSET", DEFAULT_CHARSET },
		{ "EASTEUROPE_CHARSET", EASTEUROPE_CHARSET },
		{ "GB2312_CHARSET", GB2312_CHARSET },
		{ "GREEK_CHARSET", GREEK_CHARSET },
		{ "HANGUL_CHARSET", HANGUL_CHARSET },
		{ "MAC_CHARSET", MAC_CHARSET },
		{ "RUSSIAN_CHARSET", RUSSIAN_CHARSET },
		{ "SHIFTJIS_CHARSET", SHIFTJIS_CHARSET },
		{ "SYMBOL_CHARSET", SYMBOL_CHARSET },
		{ "TURKISH_CHARSET", TURKISH_CHARSET },
		{ "VIETNAMESE_CHARSET", VIETNAMESE_CHARSET },
		{ "JOHAB_CHARSET", JOHAB_CHARSET },
		{ "ARABIC_CHARSET", ARABIC_CHARSET },
		{ "HEBREW_CHARSET", HEBREW_CHARSET },
		{ "THAI_CHARSET", THAI_CHARSET },
		{ "OEM_CHARSET", OEM_CHARSET },
		{ 0, 0 }
	};

	//Create Dummy Default
	bool bDefaultPaintHTMLTextCreated = false;
	if ( pDefaultPaintHTMLText == NULL )
	{
		bDefaultPaintHTMLTextCreated = true;
		if ( pPaintTextDefaults == NULL )
			pDefaultPaintHTMLText = new PaintHTMLText();
		else
			pDefaultPaintHTMLText = new PaintHTMLText(pPaintTextDefaults);
	}

	m_bUseAlphaMap = (bUseAlphaMap ? true : false);
	
	m_bFontVisible = iGetx2RCBool(name, (prefix + "FontVisible"), pDefaultPaintHTMLText->m_bFontVisible);
	if ( m_bFontVisible != pDefaultPaintHTMLText->m_bFontVisible )
		bDifferenceToDefault = true;

	m_iFontAlphaTransparency = iGetx2RCInt(name, (prefix + "FontAlphaTransparency"), pDefaultPaintHTMLText->m_iFontAlphaTransparency, 0, 255);
	if ( m_iFontAlphaTransparency != pDefaultPaintHTMLText->m_iFontAlphaTransparency)
		bDifferenceToDefault = true;

	m_strBorders = iGetx2RCLine(name, (prefix + "FontBorders"), pDefaultPaintHTMLText->m_strBorders);
	if ( _stricmp(m_strBorders.c_str(), pDefaultPaintHTMLText->m_strBorders.c_str()) != NULL )
		bDifferenceToDefault = true;

	if (m_strBorders.empty())
	{
		m_iLeftBorder = iGetx2RCInt(name, (prefix + "FontLeftBorder"), pDefaultPaintHTMLText->m_iLeftBorder, 0);
		m_iTopBorder = iGetx2RCInt(name, (prefix + "FontTopBorder"), pDefaultPaintHTMLText->m_iTopBorder, 0);
		m_iRightBorder = iGetx2RCInt(name, (prefix + "FontRightBorder"), pDefaultPaintHTMLText->m_iRightBorder, 0);
		m_iBottomBorder = iGetx2RCInt(name, (prefix + "FontBottomBorder"), pDefaultPaintHTMLText->m_iBottomBorder, 0);
	}
	else
	{
		Parse4SizesCFG(m_strBorders, m_iLeftBorder, m_iRightBorder, m_iTopBorder, m_iBottomBorder, true/*OnlyPositive*/);
		
		m_iLeftBorder = iGetxRCInt(name, (prefix + "FontLeftBorder"), m_iLeftBorder, 0);
		m_iTopBorder = iGetxRCInt(name, (prefix + "FontTopBorder"), m_iTopBorder, 0);
		m_iRightBorder = iGetxRCInt(name, (prefix + "FontRightBorder"), m_iRightBorder, 0);
		m_iBottomBorder = iGetxRCInt(name, (prefix + "FontBottomBorder"), m_iBottomBorder, 0);
	}

		if ( m_iLeftBorder != pDefaultPaintHTMLText->m_iLeftBorder)
			bDifferenceToDefault = true;
		if ( m_iRightBorder != pDefaultPaintHTMLText->m_iRightBorder)
			bDifferenceToDefault = true;
		if ( m_iTopBorder != pDefaultPaintHTMLText->m_iTopBorder)
			bDifferenceToDefault = true;
		if ( m_iBottomBorder != pDefaultPaintHTMLText->m_iBottomBorder)
			bDifferenceToDefault = true;

	m_bFontShadow = iGetx2RCBool(name, (prefix + "FontShadow"), pDefaultPaintHTMLText->m_bFontShadow);
	if ( m_bFontShadow != pDefaultPaintHTMLText->m_bFontShadow)
		bDifferenceToDefault = true;
	m_iFontShadowX = iGetx2RCInt(name, (prefix + "FontShadowX"), pDefaultPaintHTMLText->m_iFontShadowX);
	if ( m_iFontShadowX != pDefaultPaintHTMLText->m_iFontShadowX)
		bDifferenceToDefault = true;
	m_iFontShadowY = iGetx2RCInt(name, (prefix + "FontShadowY"), pDefaultPaintHTMLText->m_iFontShadowY);
	if ( m_iFontShadowY != pDefaultPaintHTMLText->m_iFontShadowY)
		bDifferenceToDefault = true;
	m_iFontBlockedShadow = iGetx2RCInt(name, (prefix + "FontShadowBlocked"), pDefaultPaintHTMLText->m_iFontBlockedShadow, -10, 10);
	if ( m_iFontBlockedShadow != pDefaultPaintHTMLText->m_iFontBlockedShadow)
		bDifferenceToDefault = true;
	m_clrFontShadowColor = iGetx2RCColor(name, (prefix + "FontShadowColor"), pDefaultPaintHTMLText->m_clrFontShadowColor);
	if ( m_clrFontShadowColor != pDefaultPaintHTMLText->m_clrFontShadowColor)
		bDifferenceToDefault = true;

	m_bFontOutline = iGetx2RCBool(name, (prefix + "FontOutline"), pDefaultPaintHTMLText->m_bFontOutline);
	if ( m_bFontOutline != pDefaultPaintHTMLText->m_bFontOutline)
		bDifferenceToDefault = true;
	m_clrFontOutlineColor = iGetx2RCColor(name, (prefix + "FontOutlineColor"), pDefaultPaintHTMLText->m_clrFontOutlineColor);
	if ( m_clrFontOutlineColor != pDefaultPaintHTMLText->m_clrFontOutlineColor)
		bDifferenceToDefault = true;
	m_bFontOutlineExtra = iGetx2RCBool(name, (prefix + "FontOutlineExtra"), pDefaultPaintHTMLText->m_bFontOutlineExtra);
	if ( m_bFontOutlineExtra != pDefaultPaintHTMLText->m_bFontOutlineExtra)
		bDifferenceToDefault = true;

	m_clrFontEmbossColor1 = iGetx2RCColor(name, (prefix + "FontEmbossColor1"), pDefaultPaintHTMLText->m_clrFontEmbossColor1);
	if ( m_clrFontEmbossColor1 != pDefaultPaintHTMLText->m_clrFontEmbossColor1)
		bDifferenceToDefault = true;
	m_clrFontEmbossColor2 = iGetx2RCColor(name, (prefix + "FontEmbossColor2"), m_clrFontEmbossColor1);
	if ( m_clrFontEmbossColor2 != pDefaultPaintHTMLText->m_clrFontEmbossColor2)
		bDifferenceToDefault = true;

	m_strFontName = iGetx2RCString(name, (prefix + "Font"), pDefaultPaintHTMLText->m_strFontName);
	if ( _stricmp(m_strFontName.c_str(), pDefaultPaintHTMLText->m_strFontName.c_str()) != NULL )
		bDifferenceToDefault = true;
	m_iFontHeight = iGetx2RCInt(name, (prefix + "FontHeight"), pDefaultPaintHTMLText->m_iFontHeight, 0);
	if ( m_iFontHeight != pDefaultPaintHTMLText->m_iFontHeight)
		bDifferenceToDefault = true;
	m_clrFontColor = iGetx2RCColor(name, (prefix + "FontColor"), pDefaultPaintHTMLText->m_clrFontColor);
	if ( m_clrFontColor != pDefaultPaintHTMLText->m_clrFontColor)
		bDifferenceToDefault = true;

	m_shFontCharSet = iGetx2RCNamedValue(name, (prefix + "FontCharSet"), CharSetConstants, pDefaultPaintHTMLText->m_shFontCharSet );
	if ( m_shFontCharSet != pDefaultPaintHTMLText->m_shFontCharSet)
		bDifferenceToDefault = true;

	m_bFontSmoothing = iGetx2RCBool(name, (prefix + "FontSmoothing"), pDefaultPaintHTMLText->m_bFontSmoothing);
	if ( m_bFontSmoothing != pDefaultPaintHTMLText->m_bFontSmoothing)
		bDifferenceToDefault = true;
	m_bFontClearType = iGetx2RCBool(name, (prefix + "FontClearType"), pDefaultPaintHTMLText->m_bFontClearType);
	if ( m_bFontClearType != pDefaultPaintHTMLText->m_bFontClearType)
		bDifferenceToDefault = true;

	m_strFontFadeCFG = iGetx2RCLine(name, (prefix + "FontTextFade"), pDefaultPaintHTMLText->m_strFontFadeCFG);
	if ( _stricmp(m_strFontFadeCFG.c_str(), pDefaultPaintHTMLText->m_strFontFadeCFG.c_str()) != NULL )
		bDifferenceToDefault = true;

	if (m_strFontFadeCFG.empty())
	{
		m_iLeftFade = iGetx2RCInt(name, (prefix + "FontLeftTextFade"), pDefaultPaintHTMLText->m_iLeftFade, 0);
		m_iTopFade = iGetx2RCInt(name, (prefix + "FontTopTextFade"), pDefaultPaintHTMLText->m_iTopFade, 0);
		m_iRightFade = iGetx2RCInt(name, (prefix + "FontRightTextFade"), pDefaultPaintHTMLText->m_iRightFade, 0);
		m_iBottomFade = iGetx2RCInt(name, (prefix + "FontBottomTextFade"), pDefaultPaintHTMLText->m_iBottomFade, 0);
	}
	else
	{
		m_bFontFade = Parse4SizesCFG(m_strFontFadeCFG, m_iLeftFade, m_iRightFade, m_iTopFade, m_iBottomFade, true/*OnlyPositive*/);
		
		m_iLeftFade = iGetxRCInt(name, (prefix + "FontLeftTextFade"), m_iLeftFade, 0);
		m_iTopFade = iGetxRCInt(name, (prefix + "FontTopTextFade"), m_iTopFade, 0);
		m_iRightFade = iGetxRCInt(name, (prefix + "FontRightTextFade"), m_iRightFade, 0);
		m_iBottomFade = iGetxRCInt(name, (prefix + "FontBottomTextFade"), m_iBottomFade, 0);
	}

		if ( m_iLeftFade != pDefaultPaintHTMLText->m_iLeftFade )
			bDifferenceToDefault = true;
		if ( m_iTopFade != pDefaultPaintHTMLText->m_iTopFade )
			bDifferenceToDefault = true;
		if ( m_iRightFade != pDefaultPaintHTMLText->m_iRightFade )
			bDifferenceToDefault = true;
		if ( m_iBottomFade != pDefaultPaintHTMLText->m_iBottomFade )
			bDifferenceToDefault = true;

	m_bFontBold = iGetx2RCBool(name, (prefix + "FontBold"), pDefaultPaintHTMLText->m_bFontBold);
	if ( m_bFontBold != pDefaultPaintHTMLText->m_bFontBold)
		bDifferenceToDefault = true;
	m_bFontItalic = iGetx2RCBool(name, (prefix + "FontItalic"), pDefaultPaintHTMLText->m_bFontItalic);
	if ( m_bFontItalic != pDefaultPaintHTMLText->m_bFontItalic)
		bDifferenceToDefault = true;
	m_bFontUnderline = iGetx2RCBool(name, (prefix + "FontUnderline"), pDefaultPaintHTMLText->m_bFontUnderline);
	if ( m_bFontUnderline != pDefaultPaintHTMLText->m_bFontUnderline)
		bDifferenceToDefault = true;

	m_iFontAlign = iGetx2RCNamedValue(name, (prefix + "FontAlign"), alignValues, pDefaultPaintHTMLText->m_iFontAlign);
	if ( m_iFontAlign != pDefaultPaintHTMLText->m_iFontAlign)
		bDifferenceToDefault = true;
	m_iFontVertAlign = iGetx2RCNamedValue(name, (prefix + "FontVertAlign"), vertAlignValues, pDefaultPaintHTMLText->m_iFontVertAlign);
	if ( m_iFontVertAlign != pDefaultPaintHTMLText->m_iFontVertAlign)
		bDifferenceToDefault = true;
	m_bFontNoEllipsis = iGetx2RCBool(name, (prefix + "FontNoEllipsis"), pDefaultPaintHTMLText->m_bFontNoEllipsis);
	if ( m_bFontNoEllipsis != pDefaultPaintHTMLText->m_bFontNoEllipsis)
		bDifferenceToDefault = true;

	m_bLineBreak = iGetx2RCBool(name, (prefix + "FontAutoLineBreak"), pDefaultPaintHTMLText->m_bLineBreak);
	if ( m_bLineBreak != pDefaultPaintHTMLText->m_bLineBreak)
		bDifferenceToDefault = true;
	m_iNewlineSpace = iGetx2RCInt(name, (prefix + "FontNewLineSpace"), pDefaultPaintHTMLText->m_iNewlineSpace);
	if ( m_iNewlineSpace != pDefaultPaintHTMLText->m_iNewlineSpace)
		bDifferenceToDefault = true;

	if ( bDefaultPaintHTMLTextCreated == true )
		delete pDefaultPaintHTMLText;
	
	HTMLFont.SetName(m_strFontName.c_str());
	HTMLFont.SetCharSet(m_shFontCharSet);
	HTMLFont.SetSize(m_iFontHeight);
	HTMLFont.SetBold(m_bFontBold);
	HTMLFont.SetItalic(m_bFontItalic);
	HTMLFont.SetUnderline(m_bFontUnderline);
	HTMLFont.SetSmooth(m_bFontSmoothing);
	HTMLFont.SetClearType(m_bFontClearType);
	
	return bDifferenceToDefault;
}

void PaintHTMLText::apply(HDC desthDC, int x, int y, int width, int height, LPCSTR drawText, LPSTR HREFActions, int iSetAlign, int iSetVertAlign, BOOL bSetScrolling, BOOL bDisableExtras)
{
	if ( width <= 0 || height <= 0 || !m_bFontVisible )
		return;

	RECT drawRect;
	drawRect.left = x + m_iLeftBorder;
	drawRect.top = y + m_iTopBorder;
	drawRect.right = x + width - m_iRightBorder;
	drawRect.bottom = y + height - m_iBottomBorder;

	if ( drawRect.right <= drawRect.left || drawRect.bottom <= drawRect.top )
		return;

	StringList listHREFClickRegions;

	if (bDisableExtras || (!m_bUseAlphaMap && !m_bFontFade && m_iFontAlphaTransparency == 255 && m_clrFontEmbossColor1 == RGB(255,0, 255)))
	{
		iapply(desthDC, x, y, width, height, drawText, listHREFClickRegions, iSetAlign, iSetVertAlign, (bSetScrolling ? true : false));
		
		//Prepare listHREFActions
		if (HREFActions != NULL)
		{
			strcpy(HREFActions, "");

			for (StringListIterator sit = listHREFClickRegions.begin(); sit != listHREFClickRegions.end(); sit++ )
			{
				strcat(HREFActions, (*sit).c_str());
				strcat(HREFActions, "@?@");
			}
		}
		
		return;
	}

	StringList listHREFClickRegionsTemporary;

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HDC darkDC = CreateCompatibleDC(NULL);

	HBITMAP darkBitmap = CreateDIBSection( darkDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP olddarkBitmap = (HBITMAP) SelectObject(darkDC, darkBitmap);

	HDC lightDC = CreateCompatibleDC(NULL);

	HBITMAP lightBitmap = CreateDIBSection( lightDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP oldlightBitmap = (HBITMAP) SelectObject(lightDC, lightBitmap);

	BITMAP bm_dark;
	GetObject(darkBitmap, sizeof(BITMAP), &bm_dark);

	BYTE *pPixel_dark = (LPBYTE)bm_dark.bmBits;

	int iSize = width * height;

	for (int j = 0; j < iSize; ++j)
	{
		pPixel_dark[0] = 0;
		pPixel_dark[1] = 0;
		pPixel_dark[2] = 0;
		pPixel_dark[3] = 255;

		pPixel_dark+= 4;
	}

	BITMAP bm_light;
	GetObject(lightBitmap, sizeof(BITMAP), &bm_light);

	BYTE *pPixel_light = (LPBYTE)bm_light.bmBits;

	for (j = 0; j < iSize; ++j)
	{
		pPixel_light[0] = 255;
		pPixel_light[1] = 255;
		pPixel_light[2] = 255;
		pPixel_light[3] = 255;

		pPixel_light+= 4;
	}

	iapply(darkDC, 0, 0, width, height, drawText, listHREFClickRegions, iSetAlign, iSetVertAlign, (bSetScrolling ? true : false));
	iapply(lightDC, 0, 0, width, height, drawText, listHREFClickRegionsTemporary, iSetAlign, iSetVertAlign, (bSetScrolling ? true : false));

	GetObject(darkBitmap, sizeof(BITMAP), &bm_dark);
	pPixel_dark = (LPBYTE)bm_dark.bmBits;

	GetObject(lightBitmap, sizeof(BITMAP), &bm_light);
	pPixel_light = (LPBYTE)bm_light.bmBits;

	int tmpx = 0;
	int tmpy = 0;
	int xval = 255;
	int yval = 255;

	int iLeftFade = (m_iLeftFade != 0 ? 255/m_iLeftFade : 0);
	int iRightFade = (m_iRightFade != 0 ? 255/m_iRightFade : 0);
	int iTopFade = (m_iTopFade != 0 ? 255/m_iTopFade : 0);
	int iBottomFade = (m_iBottomFade != 0 ? 255/m_iBottomFade : 0);

	for (j = 0; j < iSize; ++j)
	{
		if (pPixel_dark[3] == 0)
		{
			xval = 255;
			yval = 255;

			if ( m_bFontFade )
			{
				tmpx = (j+1)%width;
				tmpy = j/width;

				if ( tmpx == 0 )
					tmpx = width-1;

				if ( m_iLeftFade != 0 && tmpx > m_iLeftBorder && tmpx <= m_iLeftBorder+m_iLeftFade )
					xval = iLeftFade*(tmpx-m_iLeftBorder);
				else if ( m_iRightFade != 0 && tmpx > width-m_iRightBorder-m_iRightFade && tmpx <= width-m_iRightBorder )
					xval = iRightFade*(width-m_iRightBorder-tmpx);
				
				//Order is reverted Top<->Bottom
				if ( m_iBottomFade != 0 && tmpy > m_iBottomBorder && tmpy < m_iBottomBorder+m_iBottomFade )
					yval = iBottomFade*(tmpy-m_iBottomBorder);
				else if ( m_iTopFade != 0 && tmpy >= height-m_iTopBorder-m_iTopFade && tmpy < height-m_iTopBorder )
					yval = iTopFade*(height-m_iTopBorder-tmpy);
				
				if ( xval == 255 )
					pPixel_dark[3] = yval;
				else if ( yval == 255 )
					pPixel_dark[3] = xval;
				else
					pPixel_dark[3] = (xval*yval)/255;

				pPixel_dark[3] = pPixel_dark[3] - ((pPixel_light[0] - pPixel_dark[0]) * pPixel_dark[3] / 255);
			}
			else
			{
				if (pPixel_dark[0] == pPixel_light[0] && pPixel_dark[1] == pPixel_light[1] && pPixel_dark[2] == pPixel_light[2])
					pPixel_dark[3] = 255;
				else if (pPixel_dark[0] == 0 && pPixel_dark[1] == 0 && pPixel_dark[2] == 0 && pPixel_light[0] == 255 && pPixel_light[1] == 255 && pPixel_light[2] == 255 )
					pPixel_dark[3] = 0;
				else
					pPixel_dark[3] = 255 - (pPixel_light[0] - pPixel_dark[0]);
			}

			if ( pPixel_dark[0] == 255 && pPixel_dark[1] == 0 && pPixel_dark[2] == 255 )
			{
				pPixel_dark[0] = 0;
				pPixel_dark[1] = 0;
				pPixel_dark[2] = 0;
				pPixel_dark[3] = 0;
			}
			else
			{
				//PreMultiply
				pPixel_dark[0] = pPixel_dark[0] * pPixel_dark[3] / 255;
				pPixel_dark[1] = pPixel_dark[1] * pPixel_dark[3] / 255;
				pPixel_dark[2] = pPixel_dark[2] * pPixel_dark[3] / 255;
			}
		}
		else
		{
			pPixel_dark[0] = 0;
			pPixel_dark[1] = 0;
			pPixel_dark[2] = 0;
			pPixel_dark[3] = 0;
		}
					
		pPixel_dark+= 4;	
		pPixel_light+= 4;	
	}

	BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, m_iFontAlphaTransparency, AC_SRC_ALPHA };

	AlphaBlend(desthDC, x, y, width, height, darkDC, 0, 0, width, height, blendFunction);

	darkBitmap = (HBITMAP) SelectObject(darkDC, olddarkBitmap);
	DeleteDC(darkDC);
	DeleteObject(darkBitmap);
	DeleteObject(olddarkBitmap);

	lightBitmap = (HBITMAP) SelectObject(lightDC, oldlightBitmap);
	DeleteDC(lightDC);
	DeleteObject(lightBitmap);
	DeleteObject(oldlightBitmap);

	//Prepare listHREFActions
	if (HREFActions != NULL)
	{
		strcpy(HREFActions, "");

		for (StringListIterator sit = listHREFClickRegions.begin(); sit != listHREFClickRegions.end(); sit++ )
		{
			strcat(HREFActions, (*sit).c_str());
			strcat(HREFActions, "@?@");
		}
	}
}

void PaintHTMLText::iapply(HDC desthDC, int x, int y, int width, int height, const string &drawText, list<string> &listHREFClickRegions, int iSetAlign, int iSetVertAlign, bool bSetScrolling)
{
	CRect drawRect;

	if ( m_bFontNoEllipsis )
		bSetScrolling = true;

	if ( iSetAlign == -1 )
		iSetAlign = m_iFontAlign;

	if ( iSetVertAlign == -1 )
		iSetVertAlign = m_iFontVertAlign;

	drawRect.left = x + m_iLeftBorder;
	drawRect.top = y + m_iTopBorder;
	drawRect.right = x + width - m_iRightBorder;
	drawRect.bottom = y + height - m_iBottomBorder;

	//Convert into MFC variables
	CDC destCDC;
	destCDC.Attach(desthDC);

	CString strText = drawText.c_str();

	if ( strText.Find("<br") == 0 || strText.Find("< br") == 0 )
		strText.Insert(0, " ");
	else if ( strText.Find("<p") == 0 || strText.Find("< p") == 0 )
		strText.Insert(0, " ");
	strText.Replace("><br", "> <br");
	strText.Replace(">< br", "> < br");
	strText.Replace("><p", "> <p");
	strText.Replace(">< p", "> < p");

	if (m_clrFontEmbossColor1 != RGB(255,0,255))
	{
		HTMLFont.SetColor(m_clrFontEmbossColor1);

		CRect r;
		CopyRect(&r, &drawRect);
	
		r.OffsetRect(0, -1);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);

		r.OffsetRect(-1, 0);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);	
		
		r.OffsetRect(0, 1);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);

		HTMLFont.SetColor(m_clrFontEmbossColor2);
		r.OffsetRect(1, 1);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);	
		
		r.OffsetRect(1, 0);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);	
		
		r.OffsetRect(0, -1);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);

		HTMLFont.SetSmooth(false);

		HTMLFont.SetColor(RGB(255,0,255));
		r.OffsetRect(-1, 0);
		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								r,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								true);	

		HTMLFont.SetSmooth(m_bFontSmoothing);
	}
	else
	{
		if ( m_iFontBlockedShadow != 0 && !m_bFontOutline )
		{
			HTMLFont.SetColor(m_clrFontShadowColor);

			CRect r;
			CopyRect(&r, &drawRect);
			for (int i = m_iFontBlockedShadow-1; i >= 0; i--)
			{
				r.OffsetRect(-1, 1);

				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);	
			}
		}
		else if (m_bFontShadow && !m_bFontOutline)
		{
			HTMLFont.SetColor(m_clrFontShadowColor);

			CRect r;
			CopyRect(&r, &drawRect);
			//Right
			r.OffsetRect(m_iFontShadowX, m_iFontShadowY);

			CHTMLDrawer::DrawText(&destCDC,
									strText,
									HTMLFont,
									r,
									iSetAlign,
									iSetVertAlign,
									m_bLineBreak,
									m_iNewlineSpace,
									bSetScrolling,
									listHREFClickRegions,
									true);
		}
		else if (m_bFontOutline)
		{
			HTMLFont.SetColor(m_clrFontOutlineColor);

			if ( !m_bFontOutlineExtra )
			{
				CRect r;
				CopyRect(&r, &drawRect);
				//Right
				r.OffsetRect(1, 0);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);
			
				//Left
				r.OffsetRect(-2, 0);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				//(Center/)Bottom
				r.OffsetRect(1, -1);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				//Top
				r.OffsetRect(0, 2);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);
			}
			else
			{
				CRect r;
				CopyRect(&r, &drawRect);
				
				r.OffsetRect(1, 1);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(-1, 0);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(-1, 0);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(0, -1);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(0, -1);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(1, 0);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(1, 0);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);

				r.OffsetRect(0, 1);
				CHTMLDrawer::DrawText(&destCDC,
										strText,
										HTMLFont,
										r,
										iSetAlign,
										iSetVertAlign,
										m_bLineBreak,
										m_iNewlineSpace,
										bSetScrolling,
										listHREFClickRegions,
										true);
			}
		}

		drawRect.left = x + m_iLeftBorder;
		drawRect.top = y + m_iTopBorder;
		drawRect.right = x + width - m_iRightBorder;
		drawRect.bottom = y + height - m_iBottomBorder;

		HTMLFont.SetColor(m_clrFontColor);

		CHTMLDrawer::DrawText(&destCDC,
								strText,
								HTMLFont,
								drawRect,
								iSetAlign,
								iSetVertAlign,
								m_bLineBreak,
								m_iNewlineSpace,
								bSetScrolling,
								listHREFClickRegions,
								false);
	}

	destCDC.Detach();
}

POINT PaintHTMLText::measure(HDC desthDC, int x, int width, LPCSTR drawText, BOOL bWithBorders)
{
	//Convert into MFC variables
	CDC destCDC;
	destCDC.Attach(desthDC);

	CString strText = drawText;

	if ( strText.Find("<br") == 0 || strText.Find("< br") == 0 )
		strText.Insert(0, " ");
	else if ( strText.Find("<p") == 0 || strText.Find("< p") == 0 )
		strText.Insert(0, " ");
	strText.Replace("><br", "> <br");
	strText.Replace(">< br", "> < br");
	strText.Replace("><p", "> <p");
	strText.Replace(">< p", "> < p");

	CRect rctPosition; 
	if (m_bLineBreak)
	{
		rctPosition.left = x + m_iLeftBorder;
		rctPosition.right = x + width - m_iRightBorder;
	}
	else
	{
		rctPosition.left = 0;
		rctPosition.right = 0;
	}
	rctPosition.top = 0;
	rctPosition.bottom = 0;

	// Call to CHTMLDrawer
	POINT returnValues = CHTMLDrawer::MeasureText(&destCDC,
													strText,
													HTMLFont,
													rctPosition,
													m_iFontAlign,
													m_iFontVertAlign,
													m_bLineBreak,
													m_iNewlineSpace);

	destCDC.Detach();

	if (!bWithBorders)
	{
		return returnValues;
	}
	else
	{
		returnValues.x = returnValues.x+m_iLeftBorder+m_iRightBorder;
		returnValues.y = returnValues.y+m_iTopBorder+m_iBottomBorder;

		return returnValues;
	}
}