/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "PaintIcon.h"

#include "SettingsHelpers.h"
#include "PaintHelpers.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

PaintIcon::PaintIcon() :
	m_bUseAlphaMap(false),
		
	m_bIcon_Visible(true),

	m_strIcon_X(""),
	m_strIcon_Y(""),

	m_iIcon_Size(16),

	m_iIcon_AlphaTransparency(255),

	m_clrIcon_HueColor(RGB(255,255,255)),
	m_iIcon_HueIntensity(0),
	m_iIcon_LuminanceIntensity(0),
	m_iIcon_SaturationIntensity(100),

	m_clrIcon_MixColor(RGB(255,255,255)),
	m_iIcon_MixIntensity(0)
{
}

PaintIcon::PaintIcon(const PaintIconDefaults *pPaintIconDefaults) :
	m_bUseAlphaMap(false)
{
	m_bIcon_Visible = (pPaintIconDefaults->m_bIcon_Visible ? true : false);

	m_strIcon_X = pPaintIconDefaults->m_strIcon_X;
	m_strIcon_Y = pPaintIconDefaults->m_strIcon_Y;

	m_iIcon_Size = pPaintIconDefaults->m_iIcon_Size;

	m_iIcon_AlphaTransparency = pPaintIconDefaults->m_iIcon_AlphaTransparency;

	m_clrIcon_HueColor = pPaintIconDefaults->m_clrIcon_HueColor;
	m_iIcon_HueIntensity = pPaintIconDefaults->m_iIcon_HueIntensity;
	m_iIcon_LuminanceIntensity = pPaintIconDefaults->m_iIcon_LuminanceIntensity;
	m_iIcon_SaturationIntensity = pPaintIconDefaults->m_iIcon_SaturationIntensity;

	m_clrIcon_MixColor = pPaintIconDefaults->m_clrIcon_MixColor;
	m_iIcon_MixIntensity = pPaintIconDefaults->m_iIcon_MixIntensity;
}

PaintIcon::~PaintIcon()
{
}

BOOL PaintIcon::configure(LPCSTR pszName, LPCSTR pszPrefix, BOOL bUsePositioning, BOOL bUseAlphaMap, const PaintIcon *pDefaultPaintIcon, const PaintIconDefaults *pPaintIconDefaults)
{
	string name(pszName);
	string prefix(pszPrefix);

	string strAddToGroup = iGetxRCString(name, (prefix + "IconAddToGroup"), "");

	if (!strAddToGroup.empty() && _stricmp(name.c_str(), strAddToGroup.c_str()) != NULL)
	{	
		return configure(strAddToGroup.c_str(), "", bUsePositioning, bUseAlphaMap, pDefaultPaintIcon, pPaintIconDefaults);
	}

	bool bDifferenceToDefault = false;

	//Create Dummy Default
	bool bDefaultPaintIconCreated = false;
	if ( pDefaultPaintIcon == NULL )
	{
		bDefaultPaintIconCreated = true;
		if ( pPaintIconDefaults == NULL )
			pDefaultPaintIcon = new PaintIcon();
		else
			pDefaultPaintIcon = new PaintIcon(pPaintIconDefaults);
	}

	m_bUseAlphaMap = bUseAlphaMap;

	m_bIcon_Visible = iGetx2RCBool(name, (prefix + "IconVisible"), pDefaultPaintIcon->m_bIcon_Visible);
	if ( m_bIcon_Visible != pDefaultPaintIcon->m_bIcon_Visible )
		bDifferenceToDefault = true;

	m_iIcon_AlphaTransparency = iGetx2RCInt(name, (prefix + "IconAlphaTransparency"), pDefaultPaintIcon->m_iIcon_AlphaTransparency, 0, 255);
	if ( m_iIcon_AlphaTransparency != pDefaultPaintIcon->m_iIcon_AlphaTransparency )
		bDifferenceToDefault = true;

	m_iIcon_Size = iGetx2RCInt(name, (prefix + "IconSize"), pDefaultPaintIcon->m_iIcon_Size, 8, 256);
	if ( m_iIcon_Size != pDefaultPaintIcon->m_iIcon_Size )
		bDifferenceToDefault = true;

	if ( bUsePositioning )
	{
		if ( pDefaultPaintIcon->m_strIcon_X.empty() )
			m_strIcon_X = iGetx2RCString(name, (prefix + "IconX"), "0");
		else
			m_strIcon_X = iGetx2RCString(name, (prefix + "IconX"), pDefaultPaintIcon->m_strIcon_X);

		if ( pDefaultPaintIcon->m_strIcon_Y.empty() )
			m_strIcon_Y = iGetx2RCString(name, (prefix + "IconY"), ("-" + IntToString(m_iIcon_Size/2) + "c"));
		else
			m_strIcon_Y = iGetx2RCString(name, (prefix + "IconY"), pDefaultPaintIcon->m_strIcon_Y);
	}
	else
	{
		m_strIcon_X = "0";
		m_strIcon_Y = "0";
	}
	
	if ( _stricmp(m_strIcon_X.c_str(), pDefaultPaintIcon->m_strIcon_X.c_str()) != NULL )
		bDifferenceToDefault = true;
	if ( _stricmp(m_strIcon_Y.c_str(), pDefaultPaintIcon->m_strIcon_Y.c_str()) != NULL )
		bDifferenceToDefault = true;

	m_clrIcon_HueColor = iGetx2RCColor(name, (prefix + "IconHueColor"), pDefaultPaintIcon->m_clrIcon_HueColor);
	if ( m_clrIcon_HueColor != pDefaultPaintIcon->m_clrIcon_HueColor )
		bDifferenceToDefault = true;
	m_iIcon_HueIntensity = iGetx2RCInt(name, (prefix + "IconHueIntensity"), pDefaultPaintIcon->m_iIcon_HueIntensity, 0, 100);
	if ( m_iIcon_HueIntensity != pDefaultPaintIcon->m_iIcon_HueIntensity )
		bDifferenceToDefault = true;
	
	m_iIcon_LuminanceIntensity = iGetx2RCInt(name, (prefix + "IconLuminanceIntensity"), pDefaultPaintIcon->m_iIcon_LuminanceIntensity, -100, 100);
	if ( m_iIcon_LuminanceIntensity != pDefaultPaintIcon->m_iIcon_LuminanceIntensity )
		bDifferenceToDefault = true;

	m_iIcon_SaturationIntensity = iGetx2RCInt(name, (prefix + "IconSaturationIntensity"), pDefaultPaintIcon->m_iIcon_SaturationIntensity, 0, 200);
	if ( m_iIcon_SaturationIntensity != pDefaultPaintIcon->m_iIcon_SaturationIntensity )
		bDifferenceToDefault = true;

	m_clrIcon_MixColor = iGetx2RCColor(name, (prefix + "IconMixColor"), pDefaultPaintIcon->m_clrIcon_MixColor);
		if ( m_clrIcon_MixColor != pDefaultPaintIcon->m_clrIcon_MixColor )
		bDifferenceToDefault = true;
	m_iIcon_MixIntensity = iGetx2RCInt(name, (prefix + "IconMixIntensity"), pDefaultPaintIcon->m_iIcon_MixIntensity, 0, 100);
	if ( m_iIcon_MixIntensity != pDefaultPaintIcon->m_iIcon_MixIntensity )
		bDifferenceToDefault = true;

	if ( bDefaultPaintIconCreated == true )
		delete pDefaultPaintIcon;

	return bDifferenceToDefault;
}

void PaintIcon::apply(HDC desthDC, int x, int y, int width, int height, HICON hIcon)
{
	if (hIcon == NULL || !m_bIcon_Visible || width <= 0 || height <= 0)
		return;

	int iconX = ixParseCoord(m_strIcon_X, 0, width, false); 
	int iconY = ixParseCoord(m_strIcon_Y, 0, height, false);

	if ( m_bUseAlphaMap || m_iIcon_AlphaTransparency != 255 || m_iIcon_HueIntensity != 0 || m_iIcon_LuminanceIntensity != 0 || m_iIcon_SaturationIntensity != 100 || m_iIcon_MixIntensity != 0 )
	{
		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=m_iIcon_Size;      // Set size you need
		alpha.bmiHeader.biHeight=m_iIcon_Size;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		HDC darkDC = CreateCompatibleDC(NULL);

		HBITMAP darkBitmap = CreateDIBSection( darkDC,
													&alpha,
													DIB_RGB_COLORS,
													&pvBits,
													NULL,
													0 );

		HBITMAP olddarkBitmap = (HBITMAP) SelectObject(darkDC, darkBitmap);

		HDC lightDC = CreateCompatibleDC(NULL);

		HBITMAP lightBitmap = CreateDIBSection( lightDC,
													&alpha,
													DIB_RGB_COLORS,
													&pvBits,
													NULL,
													0 );

		HBITMAP oldlightBitmap = (HBITMAP) SelectObject(lightDC, lightBitmap);

		BITMAP bm_dark;
		GetObject(darkBitmap, sizeof(BITMAP), &bm_dark);

		BYTE *pPixel_dark = (LPBYTE)bm_dark.bmBits;

		int iSize = bm_dark.bmWidth * bm_dark.bmHeight;

		for (int j = 0; j < iSize; ++j)
		{
			pPixel_dark[0] = 0;
			pPixel_dark[1] = 0;
			pPixel_dark[2] = 0;
			pPixel_dark[3] = 255;

			pPixel_dark+= 4;
		}

		BITMAP bm_light;
		GetObject(lightBitmap, sizeof(BITMAP), &bm_light);

		BYTE *pPixel_light = (LPBYTE)bm_light.bmBits;

		for (j = 0; j < iSize; ++j)
		{
			pPixel_light[0] = 255;
			pPixel_light[1] = 255;
			pPixel_light[2] = 255;
			pPixel_light[3] = 255;

			pPixel_light+= 4;
		}

		//-----------------------------------------------------------------------
		//Apply Icon with colorization
		//-----------------------------------------------------------------------
		DrawIconEx(darkDC, 0, 0, hIcon, m_iIcon_Size, m_iIcon_Size, 0, NULL, DI_NORMAL);

		DrawIconEx(lightDC, 0, 0, hIcon, m_iIcon_Size, m_iIcon_Size, 0, NULL, DI_NORMAL);
		//-----------------------------------------------------------------------

		GetObject(darkBitmap, sizeof(BITMAP), &bm_dark);
		pPixel_dark = (LPBYTE)bm_dark.bmBits;

		GetObject(lightBitmap, sizeof(BITMAP), &bm_light);
		pPixel_light = (LPBYTE)bm_light.bmBits;

		BYTE r = 0;
		BYTE g = 0;
		BYTE b = 0;

		BYTE h = 0;
		BYTE s = 0;
		BYTE l = 0;
				
		BYTE hueh = 0;
		BYTE hues = 0;
		BYTE huel = 0;

		bool needHSLConversion = false;
		bool needMixColor = false;

		int iMixIntensity = m_iIcon_MixIntensity;

		if ( m_iIcon_LuminanceIntensity != 0 || m_iIcon_SaturationIntensity != 100 )
			needHSLConversion = true;

		if ( m_iIcon_HueIntensity != 0 )
		{
			r = GetRValue(m_clrIcon_HueColor);
			g = GetGValue(m_clrIcon_HueColor);
			b = GetBValue(m_clrIcon_HueColor);

			if ( r == b && b == g )
			{
				m_iIcon_HueIntensity = 0;
			}
			else
			{
				RGB2HSL(b, g, r, hueh, hues, huel); //Switch r and b due to AlphaMap/bm_b.bmHeight * bm_b.bmWidth
				needHSLConversion = true;
			}
		}

		if ( iMixIntensity != 0 )
		{
			iMixIntensity = (int) min(255, iMixIntensity*2.55);
			needMixColor = true;
		}

		for (j = 0; j < iSize; ++j)
		{
			if (pPixel_dark[0] == pPixel_light[0] && pPixel_dark[1] == pPixel_light[1] && pPixel_dark[2] == pPixel_light[2])
				pPixel_dark[3] = 255;
			else if (pPixel_dark[0] == 0 && pPixel_dark[1] == 0 && pPixel_dark[2] == 0 && pPixel_light[0] == 255 && pPixel_light[1] == 255 && pPixel_light[2] == 255 )
				pPixel_dark[3] = 0;
			else
				pPixel_dark[3] = 255 - (pPixel_light[0] - pPixel_dark[0]);

			if ( pPixel_dark[3] != 0 )
			{
				r = pPixel_dark[0];
				g = pPixel_dark[1];
				b = pPixel_dark[2];

				if ( needHSLConversion )
				{
					//AlphaMap Colorizing
					RGB2HSL(r, g, b, h, s, l);

					if(m_iIcon_HueIntensity != 0)
						h = (BYTE) max(0, min(240, (h*(100-m_iIcon_HueIntensity)+hueh*m_iIcon_HueIntensity)/100));
						
					if(m_iIcon_LuminanceIntensity != 0)
						l = (BYTE) max(0, min(240, ((l*100)+(120*m_iIcon_LuminanceIntensity))/100));

					if(m_iIcon_SaturationIntensity != 100)
					{
						if ( m_iIcon_SaturationIntensity > 100 )
							s = (BYTE) max(0, min(240, (240*(m_iIcon_SaturationIntensity-100))/100));				
						else
							s = (BYTE) max(0, min(240, (s*m_iIcon_SaturationIntensity)/100));
					}

					HSL2RGB(r, g, b, h, s, l);
				}

				if ( needMixColor )
				{
					r = (BYTE)((r * (255 - iMixIntensity) + GetBValue(m_clrIcon_MixColor) * iMixIntensity + 255) >> 8);
					g = (BYTE)((g * (255 - iMixIntensity) + GetGValue(m_clrIcon_MixColor) * iMixIntensity + 255) >> 8);
					b = (BYTE)((b * (255 - iMixIntensity) + GetRValue(m_clrIcon_MixColor) * iMixIntensity + 255) >> 8);
				}

				pPixel_dark[0] = r;
				pPixel_dark[1] = g;
				pPixel_dark[2] = b;
			}

			//PreMultiply
			pPixel_dark[0] = pPixel_dark[0] * pPixel_dark[3] / 255;
			pPixel_dark[1] = pPixel_dark[1] * pPixel_dark[3] / 255;
			pPixel_dark[2] = pPixel_dark[2] * pPixel_dark[3] / 255;
					
			pPixel_dark+= 4;	
			pPixel_light+= 4;	
		}

		BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, m_iIcon_AlphaTransparency, AC_SRC_ALPHA };

		AlphaBlend(desthDC, x+iconX, y+iconY, m_iIcon_Size, m_iIcon_Size, darkDC, 0, 0, m_iIcon_Size, m_iIcon_Size, blendFunction);

		darkBitmap = (HBITMAP) SelectObject(darkDC, olddarkBitmap);
		DeleteDC(darkDC);
		DeleteObject(darkBitmap);
		DeleteObject(olddarkBitmap);

		lightBitmap = (HBITMAP) SelectObject(lightDC, oldlightBitmap);
		DeleteDC(lightDC);
		DeleteObject(lightBitmap);
		DeleteObject(oldlightBitmap);
	}
	else
	{
		//-----------------------------------------------------------------------
		//Apply Icon 
		//-----------------------------------------------------------------------
		DrawIconEx(desthDC, x+iconX, y+iconY, hIcon, m_iIcon_Size, m_iIcon_Size, 0, NULL, DI_NORMAL);
	}
}
