/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "PaintTooltip.h"

#include "SettingsHelpers.h"

#include "shlobj.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

PaintTooltip::PaintTooltip(HWND hWndParent, HINSTANCE hInstanceParent) :
	hWndParentWindow(NULL),
	hInstanceParentWindow(NULL),
	hWndTooltip(NULL)
{
	hWndParentWindow = hWndParent;
	hInstanceParentWindow = hInstanceParent;
}

PaintTooltip::~PaintTooltip()
{
	RECT emptyRect = {0,0,0,0};
	
	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWndParentWindow;
	ti.uId = 0;
	ti.rect = emptyRect;
	ti.hinst = hInstanceParentWindow;
	ti.lpszText = NULL;
	ti.lParam = 0;

	SendMessage(hWndTooltip, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);

	DestroyWindow(hWndTooltip);
	hWndTooltip = NULL;
}

HWND PaintTooltip::configure(LPCSTR pszName, LPCSTR pszPrefix)
{
	string name(pszName);
	string prefix(pszPrefix);

	string strAddToGroup = iGetxRCString(name, (prefix + "TooltipAddToGroup"), "");

	if (!strAddToGroup.empty() && _stricmp(name.c_str(), strAddToGroup.c_str()) != NULL)
	{	
		return configure(strAddToGroup.c_str(), "");
	}

	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";

	//ToolTips
	bool bBalloonTooltip = false;
	string strBalloonTitle = "";
	int iBalloonIcon = 0;

	numTokens = SplitString(iGetx2RCLine(name, (prefix + "TooltipBalloon"), "false"), one, two, three, four, five, six, seven);

	if ( _stricmp( one.c_str(), "false") != NULL && (IsOS(OS_2KXP) || IsOS(OS_ME)) )
	{
		bBalloonTooltip = true;

		if ( numTokens >= 1 && !one.empty() && _stricmp( one.c_str(), "true") != NULL )
		{
			strBalloonTitle = one;
			if ( numTokens >= 2 )
				iBalloonIcon = atoi( two.c_str() );
			if ( iBalloonIcon > 3 || iBalloonIcon < 0 )
				iBalloonIcon = 0;
		}
	}

	// struct specifying control classes to register
	INITCOMMONCONTROLSEX iccex; 

	/* INITIALIZE COMMON CONTROLS */
	iccex.dwICC = ICC_WIN95_CLASSES;
	iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);

	InitCommonControlsEx(&iccex);

	hWndTooltip = CreateWindowEx(WS_EX_TOOLWINDOW,
										TOOLTIPS_CLASS,
										TEXT(""),
										WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP | (bBalloonTooltip ? TTS_BALLOON : 0),
										CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
										NULL,
										NULL,
										hInstanceParentWindow,
										NULL);
	
	SetWindowPos(hWndTooltip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);  
		
	SendMessage(hWndTooltip, TTM_SETMAXTIPWIDTH, 0, iGetx2RCInt(name, (prefix + "ToolTipMaxWidth"), 300, 50));

	SendMessage(hWndTooltip, (UINT) TTM_SETTIPBKCOLOR, (WPARAM) (COLORREF) iGetx2RCColor(name, (prefix + "TooltipColor"), GetSysColor(COLOR_INFOBK)), 0 );   
	SendMessage(hWndTooltip, (UINT) TTM_SETTIPTEXTCOLOR, (WPARAM) (COLORREF) iGetx2RCColor(name, (prefix + "ToolTipTextColor"), GetSysColor(COLOR_INFOTEXT)), 0);  
			
	numTokens = SplitString(iGetx2RCLine(name, (prefix + "ToolTipDurations"), "-1 -1"), one , two, three, four, five, six, seven);
			
	int initial = -1;
	int autopop = -1;

	if ( numTokens >= 1 )
	{
		initial = atoi(one.c_str());
		if ( numTokens >= 2 )
			autopop = atoi(two.c_str());
	}

	if ( initial < 0 )
		initial = -1;
	if ( autopop < 1000 )
		autopop = 1000;

	if ( initial != -1 )
		SendMessage(hWndTooltip, (UINT) TTM_SETDELAYTIME, (WPARAM) (DWORD) TTDT_INITIAL, (LPARAM) MAKELONG(initial, 0) );
	if ( autopop != -1 )
		SendMessage(hWndTooltip, (UINT) TTM_SETDELAYTIME, (WPARAM) (DWORD) TTDT_AUTOPOP, (LPARAM) MAKELONG(autopop, 0) );

	if ( bBalloonTooltip && !strBalloonTitle.empty() )
		SendMessage(hWndTooltip, TTM_SETTITLE, (WPARAM) (int) iBalloonIcon, (LPARAM) strBalloonTitle.c_str() );

	RECT rectParent;
	GetClientRect(hWndParentWindow, &rectParent);

	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_TRANSPARENT;
	ti.hwnd = hWndParentWindow;
	ti.uId = 0;
	ti.rect = rectParent;
	ti.hinst = hInstanceParentWindow;
	ti.lpszText = "";
	ti.lParam = 0;

	SendMessage(hWndTooltip, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);

	return hWndTooltip;
}
	
void PaintTooltip::apply(LPCSTR strTooltipText, BOOL bUpdateSize)
{
	string strProcessTooltip = strTooltipText;

	if ( !strProcessTooltip.empty() )
	{
		while ( strstr(strProcessTooltip.c_str(), "<br>") != NULL )
			strProcessTooltip.replace(strProcessTooltip.find("<br>"), 4, "\n", 1);
		
		while ( strstr(strProcessTooltip.c_str(), "\\n") != NULL )
			strProcessTooltip.replace(strProcessTooltip.find("\\n"), 2, "\n", 1);
		
		while ( strstr(strProcessTooltip.c_str(), "\\t") != NULL )
			strProcessTooltip.replace(strProcessTooltip.find("\\t"), 2, "\t", 1);
	}

	char charTooltipText[MAX_LINE_LENGTH] = {0};
	strcpy(charTooltipText, strProcessTooltip.c_str());

	RECT rectParent;
	GetClientRect(hWndParentWindow, &rectParent);

	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_TRANSPARENT;
	ti.hwnd = hWndParentWindow;
	ti.uId = 0;
	ti.rect = rectParent;
	ti.hinst = hInstanceParentWindow;
	ti.lpszText = charTooltipText;
	ti.lParam = 0;

	if (bUpdateSize)
		SendMessage(hWndTooltip, TTM_NEWTOOLRECT, 0, (LPARAM) (LPTOOLINFO) &ti);
	else
		SendMessage(hWndTooltip, TTM_UPDATETIPTEXT, 0, (LPARAM) (LPTOOLINFO) &ti);
}
