/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Exported in xExports.h
//////////////////////////////////////////////////////////////////////

BOOL IsOS(const int &os)
{	
	static int osversion = 0;

	static bool bInitialized = false;

	if (!bInitialized)
	{
		OSVERSIONINFO verInfo;
		verInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		GetVersionEx(&verInfo);
		
		switch(verInfo.dwPlatformId)
		{
			case VER_PLATFORM_WIN32_WINDOWS:

				osversion |= OS_9X;
			
				if(verInfo.dwMinorVersion >= 90)
				{
					osversion |= OS_ME;
				}
				else if(verInfo.dwMinorVersion >= 10)
				{
					osversion |= OS_98;
				}
				else
				{
					osversion |= OS_95;
				}
				break;
			
			case VER_PLATFORM_WIN32_NT:
			
				osversion |= OS_NT;
			
				if(verInfo.dwMajorVersion >= 5)
				{
					osversion |= OS_2KXP;

					if (verInfo.dwMinorVersion >= 1)
					{
						osversion |= OS_XP;
					}
					else
					{
						osversion |= OS_2K;
					}
				}
				else
				{
					osversion |= OS_NT4;
				}
				break;
			default:;
		}
		bInitialized = true;
	}

	return (osversion & os) ? true : false;
}

int QueryPixelDepth()
{
	DEVMODE DevMode;
	DevMode.dmSize=sizeof(DEVMODE);
	
	if (EnumDisplaySettings(NULL,ENUM_CURRENT_SETTINGS,&DevMode))
		return (int) DevMode.dmBitsPerPel;
	else
		return 8;
}

// retrieve handle of the desktop window
HWND GetLitestepDesktop()
{
	HWND hWnd = FindWindow("DesktopBackgroundClass", NULL);

	if(hWnd)
		return hWnd;

	return GetDesktopWindow();
}

// is the given window a toplevel application window?
BOOL IsAppWindow(const HWND &hWnd)
{
	if (!IsWindow(hWnd))
		return false;
	
	if (!IsWindowVisible(hWnd))
		return false;

	if ( GetWindowLong(hWnd, GWL_USERDATA) == MAGIC_DWORD )
		return false;

	long exStyle = GetWindowLong(hWnd, GWL_EXSTYLE);

	if ( exStyle & WS_EX_APPWINDOW )
		return true;
	
	if ( GetParent( hWnd ) )
		return false;
	
	if ( GetWindow(hWnd, GW_OWNER) )
		return false;

	if ( exStyle & WS_EX_TOOLWINDOW )
		return false;

	return true;
}

// modify window styles
void ModifyStyle(HWND hWnd, DWORD removeStyle, DWORD addStyle)
{
	SetWindowLong(hWnd, GWL_STYLE, (GetWindowLong(hWnd, GWL_STYLE) & ~removeStyle) | addStyle);
}

void ModifyStyleEx(HWND hWnd, DWORD removeStyle, DWORD addStyle)
{
	SetWindowLong(hWnd, GWL_EXSTYLE, (GetWindowLong(hWnd, GWL_EXSTYLE) & ~removeStyle) | addStyle);
}
